/* Simulator for pipelined implementation of MIPS */
#include <stdio.h>
#include "pipe.h"
#include "mips.h"

/* Run time flags */
/* Allow processor to signal when it has encountered halt */
int break_encountered = 0;
/* Should I be verbose in printing the state of the simulation? */
int verbose = 1;
FILE *outfile = stdout;

/* Define states held in pipe register stages */

/* Program counter */
typedef struct {
  unsigned pc;
} pc_ele, *pc_ptr;

/* IF/ID pipe register */
typedef struct {
  unsigned instr;
  unsigned incr_pc;
} if_id_ele, *if_id_ptr;

/* ID/EX Pipe Register */
typedef struct {
  unsigned incr_pc;
  unsigned instr;
  unsigned adata;
  unsigned bdata;
} id_ex_ele, *id_ex_ptr;

/* EX/MEM Pipe Register */
typedef struct {
  unsigned branch_target;
  unsigned char branch_flag;
  unsigned char mem_write_flag;
  unsigned char mem_read_flag;
  unsigned alu_result;
  unsigned bdata;
  unsigned char write_back_dest;
} ex_mem_ele, *ex_mem_ptr;

/* MEM/WB Pipe Register */
typedef struct {
  unsigned char mem_read_flag;
  unsigned char write_back_dest;
  unsigned alu_result;
  unsigned read_data;
} mem_wb_ele, *mem_wb_ptr;

/*** Register File ****/

/* Register file contents */
unsigned reg_state[32] = {0};

/* Perform one write and two reads to register file */
void reg_update(unsigned char Waddr, unsigned Wdata,
		unsigned char Aaddr, unsigned char Baddr,
		unsigned *Adata, unsigned *Bdata)
{
  /* First do the write */
  if (Waddr >= 32)
    fprintf(outfile, "WARNING: Attempt to write register %u\n", Waddr);
  else
    if (Waddr > 0)
      reg_state[Waddr] = Wdata;
  /* Now do the reads */
  if (Aaddr >= 32)
    fprintf(outfile, "WARNING: Attempt to read register %u\n", Aaddr);
  else
    *Adata = reg_state[Aaddr];
  if (Baddr >= 32)
    fprintf(outfile, "WARNING: Attempt to read register %u\n", Baddr);
  else
    *Bdata = reg_state[Baddr];
}

/****************** Memories ******************************/

#define MEM_COUNT 128
unsigned instr_mem[MEM_COUNT] = {0};
/* Set of strings describing each of the instructions */
char *instr_code[MEM_COUNT] = {0};
unsigned data_mem[MEM_COUNT] = {0}; 

unsigned instr_fetch(unsigned pc)
{
  unsigned instr;
  char *code;
  if (pc & 0x3) {
    fprintf(outfile, "WARNING: Unaligned instruction fetch %u\n", pc); 
  }
  pc >>= 2;
  if (pc >= MEM_COUNT) {
    fprintf(outfile, "WARNING: Attempt to read instruction address %u\n", pc);
    return BREAK_INSTR(0);
  }
  instr = instr_mem[pc];
  code = instr_code[pc];
  if (verbose)
    fprintf(outfile, "Fetching instruction %x:0x%x %s\n",
	    pc, instr, code);
  return instr;
}

unsigned data_read(unsigned addr)
{
  unsigned result;
  if (addr & 0x3) {
    fprintf(outfile, "WARNING: Unaligned read %u\n", addr); 
  }
  addr >>= 2;
  if (addr >= MEM_COUNT) {
    fprintf(outfile, "WARNING: Attempt to read memory address %u\n", addr);
    result = 0;
  }
  else
    result = data_mem[addr];
  if (verbose)
    fprintf(outfile, "Reading word 0x%x:0x%x\n", addr<<2, result);
  return result;
}

void data_write(unsigned addr, unsigned data)
{
  if (addr & 0x3) {
    fprintf(outfile, "WARNING: Unaligned write %u\n", addr); 
  }
  addr >>= 2;
  if (addr >= MEM_COUNT)
    fprintf(outfile, "WARNING: Attempt to write memory address %u\n", addr);
  else {
    data_mem[addr] = data;
    if (verbose)
      fprintf(outfile, "Writing word 0x%x:0x%x\n", addr<<2, data);
  }
}

/**************** ALU ********************************/
unsigned alu_compute(unsigned dataA, unsigned dataB, unsigned char op,
		     unsigned char *is_zero_ptr)
{
  unsigned result = 0;
  switch (op) {
  case O_SLL:
    /* Don't implement actual shift */
    result = dataB;
    break;
  case O_ADDU:
    result = dataA + dataB;
    break;
  case O_NOR:
    result = ~(dataA | dataB);
    break;
  case O_OR:
    result = (dataA | dataB);
    break;
  case O_SLT:
    result = ((int) dataA < (int) dataB);
    break;
  case O_SLTU:
    result = (dataA < dataB);
    break;
  case O_SUBU:
    result = dataA - dataB;
    break;
  case O_XOR:
    result = dataA ^ dataB;
  case O_BREAK:
    result = 0;
    break;
  default:
    fprintf(outfile, "WARNING: Illegal ALU operation %u\n", op);
    break;
  }
  if (is_zero_ptr)
    *is_zero_ptr = (result == 0);
  return result;
}

/*************** Simulator state *********************/


/* The different stages */
stage_ptr pc_state, if_id_state, id_ex_state, ex_mem_state, mem_wb_state;

pc_ptr pc_current, pc_next;
if_id_ptr if_out, id_in;
id_ex_ptr id_out, ex_in;
ex_mem_ptr ex_out, mem_in;
mem_wb_ptr mem_out, wb_in;

void init_state()
{
  pc_state = new_stage(sizeof(pc_ele));
  pc_next = pc_state->next;
  pc_current = pc_state->current;
  if_id_state = new_stage(sizeof(if_id_ele));
  if_out = if_id_state->next;
  id_in = if_id_state->current;
  id_ex_state = new_stage(sizeof(id_ex_ele));
  id_out = id_ex_state->next;
  ex_in = id_ex_state->current;
  ex_mem_state = new_stage(sizeof(ex_mem_ele));
  ex_out = ex_mem_state->next;
  mem_in = ex_mem_state->current;
  mem_wb_state = new_stage(sizeof(mem_wb_ele));
  mem_out = mem_wb_state->next;
  wb_in = mem_wb_state->current;
  clear_state(reg_state, 32);
  clear_state(instr_mem, MEM_COUNT);
  clear_state(data_mem, MEM_COUNT);
  break_encountered = 0;
}

void do_if_stage()
{
  unsigned pc = pc_current->pc;
  unsigned pc4 = pc+4;
  if_out->instr = instr_fetch(pc);
  if_out->incr_pc = pc4;
  /* Update PC to be either branch target or incremented PC */
  pc_next->pc = mem_in->branch_flag ?
    mem_in->branch_target : pc4;
}

void do_id_stage()
{
  /* Get parameters for write back */
  unsigned char write_reg = wb_in->write_back_dest;
  unsigned write_data = wb_in->mem_read_flag ?
    wb_in->read_data : wb_in->alu_result;

  /* Get parameters from instruction */ 
  unsigned instr = id_in->instr;
  unsigned char rs = GET_RS(instr);
  unsigned char rt = GET_RT(instr);

  /* Update stage outputs */
  id_out->incr_pc = id_in->incr_pc;
  id_out->instr = instr;
  /* Do register operations */
  reg_update(write_reg, write_data, rs, rt,
	     &id_out->adata, &id_out->bdata);
}

void do_ex_stage()
{
  /* Set various parameters to most likely values */
  unsigned char zero;
  unsigned char branch_flag = 0;
  unsigned instr = ex_in->instr;
  unsigned char alu_fun = GET_FUN(instr);
  unsigned instr_op = GET_OPCODE(instr);
  unsigned char mem_read_flag = 0;
  unsigned char mem_write_flag = 0;
  unsigned char branch_if_zero = 0;
  unsigned char branch_if_nonzero = 0;
  /* Operands for ALU. */
  unsigned alu_adata = ex_in->adata;
  unsigned alu_bdata = ex_in->bdata;
  /* Bdata passed to MEM */
  unsigned bdata = alu_bdata;

  unsigned char rt = GET_RT(instr);
  unsigned char write_back_dest = GET_RD(instr);
  short i16 = GET_I16(instr);
  /* Compute branch target */
  ex_out->branch_target = ex_in->incr_pc
    + (((int) i16) << 2); /* Sign extend */
  switch (instr_op) {
  case I_OP:
    break;  /* Use defaults */
  case I_ADDIU:
    alu_fun = O_ADDU;
    alu_bdata = (unsigned short) i16; /* Don't sign extend */
    write_back_dest = rt;
    break;
  case I_LW:
    alu_fun = O_ADDU;
    alu_bdata = i16; /* Sign extend */
    mem_read_flag = 1;
    write_back_dest = rt;
    break;
  case I_SW:
    alu_fun = O_ADDU;
    alu_bdata = i16; /* Sign extend */
    mem_write_flag = 1;
    write_back_dest = 0;
    break;
  case I_BEQ:
    alu_fun = O_SUBU;
    branch_if_zero = 1;
    write_back_dest = 0;
    break;
  case I_BNE:
    alu_fun = O_SUBU;
    branch_if_nonzero = 1;
    write_back_dest = 0;
    break;
  case I_LUI:
    alu_fun = O_ADDU;
    alu_bdata = i16<<16; /* Shift into upper half */
    write_back_dest = rt;
    break;
  default:
    fprintf(outfile, "WARNING: Illegal instruction encountered in EX %d\n",
	    instr_op);
    break;
  }
  /* Do the ALU operation */
  ex_out->alu_result = alu_compute(alu_adata, alu_bdata, alu_fun, &zero);
  ex_out->branch_flag = zero && branch_if_zero || !zero && branch_if_nonzero;
  ex_out->mem_write_flag = mem_write_flag;
  ex_out->mem_read_flag = mem_read_flag;
  ex_out->write_back_dest = write_back_dest;
  ex_out->bdata = bdata;
  if (verbose) 
    fprintf(outfile,
	    "EX Results: ALU:0x%x, BR:%u, MW:%u, MR:%u, WB:0x%x, B:0x%x\n",
	    ex_out->alu_result, ex_out->branch_flag, ex_out->mem_write_flag,
	    ex_out->mem_read_flag, ex_out->write_back_dest,
	    ex_out->bdata);
}

void do_mem_stage()
{
  unsigned address = mem_in->alu_result;
  unsigned data = 0;
  unsigned sdata = mem_in->bdata;
  if (mem_in->mem_read_flag)
    data = data_read(address);
  if (mem_in->mem_write_flag)
    data_write(address, sdata);
  mem_out->mem_read_flag = mem_in->mem_read_flag;
  mem_out->write_back_dest = mem_in->write_back_dest;
  mem_out->alu_result = mem_in->alu_result;
  mem_out->read_data = data;
}

/* Implementation of stalls */
/* This check is run after all the stages have been updated, so can use
   both stage outputs and inputs */
void do_stall_check()
{
  /* Should I stall the instruction in ID? */
  unsigned char stall_id = 0;
  /* Should I stall the instruction in EX? */
  unsigned char stall_ex = 0;

  /* Presently,
     stall only in the event that BREAK instruction detected in EX.
     This will need to be embellished!
  */
  if (ex_in->instr == BREAK_INSTR(0)) {
    break_encountered = 1;
    stall_id = 1;
  }
  /***** End of stall detection logic *****/
  if (stall_ex) {
    pc_state->stall = 1;
    if_id_state->stall = 1;
    id_ex_state->stall = 1;
    /* Insert bubble into MEM */
    ex_mem_state->bubble = 1;
  }
  /* Handle stall at ID */
  else if (stall_id) {
    pc_state->stall = 1;
    if_id_state->stall = 1;
    /* Insert bubble into EX */ 
    id_ex_state->bubble = 1;
  }
}

/* Run pipeline for one cycle */
void step_pipe()
{
  do_if_stage();
  do_id_stage();
  do_ex_stage();
  do_mem_stage();
  do_stall_check();
  update_stages();
}

/* Run pipeline for count cycles or until halted */
void run_pipe(int count, unsigned start_pc)
{
  unsigned cycles = 0;
  pc_current->pc = start_pc;
  break_encountered = 0;
  while (count--) {
    step_pipe();
    cycles++;
    if (verbose) {
      fprintf(outfile, "Nonzero register state after %d cycles\n", cycles);
      dump_state(reg_state, NULL, 32, outfile);
    }
    if (break_encountered)
      /* Run just enough cycles to complete instructions ahead of break */
      count = count <= 2 ? count : 2;
  }
  if (!verbose) {
    fprintf(outfile, "Nonzero register state after %d cycles\n", cycles);
    dump_state(reg_state, NULL, 32, outfile);
  }
  
}
     
/* Run the simulator */
/* Arguments should be the names of the code, cycle limit, and data files */ 
main(int argc, char *argv[])
{
  FILE *code_file, *data_file;
  unsigned code_count, data_count;
  int clock_limit = 100;
  if (argc == 1) {
    fprintf(stderr, "Usage: %s instr_file [limit [data_file]]\n", argv[0]);
    exit(0);
  }
  init_state();
  if (argc > 1) {
    code_file = fopen(argv[1], "r");
    if (!code_file) {
      fprintf(stderr, "Couldn't locate code file %s\n", argv[1]);
      exit(1);
    }
    code_count = load_state(instr_mem, instr_code, code_file);
    fclose(code_file);
    if (code_count == 0) {
      fprintf(stderr, "No lines of code found\n");
      exit(1);
    }
    if (verbose) {
      fprintf(outfile, "%u lines of code read\n", code_count);
      dump_state(instr_mem, instr_code, code_count, outfile);
    }
  }
  if (argc > 2) {
    sscanf(argv[2], "%u", &clock_limit);
  }
  if (argc > 3) {
    data_file = fopen(argv[3], "r");
    if (!data_file) {
      fprintf(stderr, "Couldn't locate data file %s\n", argv[3]);
      exit(1);
    }
    load_state(data_mem, NULL, data_file);
  }
  run_pipe(clock_limit, 0);
  fclose(outfile);
  exit(0);
}
